/*
 *  NXPostScript - nximage2ps.c
 *
 *      Version 2.0
 *      by Lance Lovette and Robert Ward
 *      Code and Display Copyright (C) 1993 by Lance Lovette
 *      All Rights Reserved
 *
 *      Lance Lovette (lovette@iftech.com)
 *	Robert Ward (ward@iftech.com)
 *      North Carolina State University
 *      Engineering Computer Operations
 *      Department of Computer Science
 *      Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of the authors and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  The authors and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  The authors AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL the authors OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nximage2ps.c,v $
 *  $Revision: 2.1 $
 *  $Date: 1993/12/10 04:14:06 $
 */
static char rcsid[] = "$Id: nximage2ps.c,v 2.1 1993/12/10 04:14:06 ltlovett Exp $";

#include "NXPostScript.h"

#define BYTES_PER_OUTPUT_LINE	36	/* The number of hex characters to print per line. */

/***********************************************************/

/*
 *  Proc:	XImageToBitString
 *  Callers:	NXPixmapToPostScript
 *  Purpose:	Converts an XImage to a string of characters.  Each byte of the string
 *		represents 8 bits in the image (a number 0 -> 255).
 *  Params:	image - Pointer to the XImage.
 *		bitStringWidth - <Return> The number of bytes per line of the image.
 *		bitStringLen - <Return> The number of bytes in the string.
 *  Globals:	None
 *  Returns:	The bit string.
 */
unsigned char *XImageToBitString( XImage *image, int *bitStringWidth, int *bitStringLen )
{
	int			y;
	unsigned char		*bitString;

	register int		x;
	register unsigned char	currentByte, currentBit;
	register unsigned char	*ptr;

	if( !image ) return( NULL );
	if( !bitStringLen ) return( NULL );

	/*
	 * The XImage structure tells us how many bytes there are in one
	 * scan line of the image.  Multiply this number by the height
	 * (number of scan lines) in the image to get the total number of
	 * bytes we will need for storage.
	 */
	(*bitStringWidth) = image->bytes_per_line;
	(*bitStringLen) = (*bitStringWidth) * image->height;

	/*
	 * Make room for the string - initialize the bytes to zero.
	 */
	bitString = (unsigned char *) XtCalloc( (*bitStringLen), sizeof(unsigned char) );
	if (!bitString) return( 0 );

	currentByte = 0; 
	currentBit = 128;
	ptr = bitString;

	/*
	 * Algorithm:
	 *
	 * Start at the MSB position in the current byte (currentBit) and
	 * set it to a 1.  Starting at the lower left corner of the image
	 * get the first pixel.  If the pixel is non-white, set that bit
	 * in the current byte (currentByte) to a 1 by or'ing.  Working
	 * left to right/bottom to top, move the bit position (currentBit)
	 * to the right (by shifting).  Only 1 bit position in currentBit will be
	 * 1 - the bit we are looking at.  After filling a byte with 8 image
	 * pixels we have to invert it because in PostScript a 1 is white.
	 * Keep shifting and or'ing until the entire image is processed.  Any
	 * bit positions that are not set will be 0 by default since we used
	 * calloc for memory allocation.
	 */
	for( y=image->height - 1; y >= 0; y-- )
	{
		for( x=0; x < image->width; )
		{
			if( XGetPixel(image, x, y) ) currentByte |= currentBit;

			++x;
			currentBit >>= 1;

			if( !(x & 7) )
			{
				*(ptr++) = ~currentByte;
				currentByte = 0; 
				currentBit = 128;
			}
		}

		if( (x & 7) )
		{
			*(ptr++) = ~currentByte;
			currentByte = 0; 
			currentBit = 128;
		}
	}

	return( bitString );
}

/***********************************************************/

/*
 *  Proc:	NXPixmapToPostScript
 *  Callers:	<Library>
 *  Purpose:	Converts a Pixmap image into a PostScript image and outputs
 *		the PostScript code that would be required to print the image.
 *		The code is not printable - it does not have a 'showpage' - it
 *		is meant to be "inlined" in other PostScript code.
 *  Params:	display - A display to work from.
 *		destFile - The destination file to output to.
 *		pixmap - The pixmap to convert.
 *		bps - Bits per sample of the pixmap (for PostScript).
 *		pixmapWidth - The actual width of the pixmap.
 *		pixmapHeight - The actual height of the pixmap.
 *		boundingBoxWidth - The width of the total area the pixmap should occupy.
 *		boundingBoxHeight - The height of the total area the pixmap should occupy.
 *  Globals:	None
 *  Returns:	0 if the conversion was successful.
 */
int NXPixmapToPostScript( Display *display, FILE *destFile, Pixmap pixmap,
				unsigned int bps, unsigned int pixmapWidth, unsigned int pixmapHeight,
				unsigned int boundingBoxWidth, unsigned int boundingBoxHeight )
{
	int		c;
	int		currentByte;
	int		bitStringLen;
	int		bitStringWidth;
	XImage		*image;
	unsigned char	*ptr;
	unsigned char	*bitString;

	if( !display ) return( 1 );
	if( !destFile ) return( 1 );

	/*
 	 * Convert the Pixmap to an XImage, then convert the XImage into a
	 * string of hexadecimal characters that can be output.
	 */
	image = XGetImage( display, pixmap, 0, 0, pixmapWidth, pixmapHeight, 1L, XYPixmap );
	bitString = XImageToBitString( image, &bitStringWidth, &bitStringLen );

	XDestroyImage (image);

	if( !bitString ) return( 1 );

	/*
	 * The model that the NXPostScript driver wants is two integers
	 * that describe the bounding box width and height.  These are used
	 * by the driver and serve no other purpose.
	 */
	fprintf( destFile, "    %d %d pop pop\n", boundingBoxWidth, boundingBoxHeight );

	fprintf( destFile, "    %% BeginImage\n" );

	/*
	 * Print out the image as a block of hexadecimal characters.
	 */
	fprintf( destFile, "\t/_NXimagedata %d string def\n", bitStringWidth );
	fprintf( destFile, "\t%d %d %d [ %d 0 0 %d 0 0 ]\n", pixmapWidth, pixmapHeight, bps, pixmapWidth, pixmapHeight );
	fprintf( destFile, "\t{ currentfile _NXimagedata readhexstring not { quit } if } bind image\n" );
	fprintf( destFile, "\t\t" );

	for( currentByte=0, ptr=bitString; currentByte < bitStringLen; currentByte++, ptr++ )
	{
		if( currentByte && !(currentByte % BYTES_PER_OUTPUT_LINE) ) fprintf (destFile, "\n\t\t");

		c = (int)(*ptr);
		if ( c < 0 ) c += 256;

		fprintf (destFile, "%02x", c);
	}

	fprintf( destFile, "\n    %% EndImage\n" );

	XtFree( (char *)bitString );

	return( 0 );
}

