/*
 *  NXUtils - nxini.c 
 *
 *	Version 2.0
 *      by Lance Lovette
 *	Code and Display Copyright (C) 1993 by Lance Lovette
 *	All Rights Reserved
 *
 *	Lance Lovette (lovette@iftech.com)
 *	North Carolina State University
 *	Engineering Computer Operations
 *	Department of Computer Science
 *	Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of Lance Lovette and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  Lance Lovette and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  Lance Lovette AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Lance Lovette OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nxini.c,v $
 *  $Revision: 2.4 $
 *  $Date: 1994/01/02 20:03:51 $
 */
static char rcsid[] = "$Id: nxini.c,v 2.4 1994/01/02 20:03:51 ltlovett Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#include "NXUtils.h"

#define	INI_VERSION_STR		"1.0.0"				/* Current library version as a string. */
#define	INI_VERSION_FLT		"1.00"				/* Current library version as a float. */
#define	MAX_INI_LINE_LEN	512				/* The maximum length of an input line. */

typedef long filepos_t;

/***********************************************************************
 *
 *
 *  The following functions are private functions that are used by
 *  the public XnIni functions below.  These functions should really
 *  never be used by the common application.
 *
 *
 ***********************************************************************/

/*
 *  Proc:	IniWriteLine
 *  Callers:
 *  Purpose:	Write a line to the ini file.
 *  Params:	file - The file to write to.
 *		line - The text to write.
 *  Returns:	INI_SUCCESS if the line is successfully written.
 */
int IniWriteLine( FILE *file, String line )
{
	if( !file || !line ) return( INI_INVALID_VALUE );

	clearerr( file );

	fprintf( file, "%s", line );

	if( ferror(file) == 0 )
		return( INI_SUCCESS );

	return( INI_INVALID_FILE );
}

/***********************************************************************/

/*
 *  Proc:	IniCreateFile
 *  Callers:
 *  Purpose:	Create an empty ini file.
 *  Params:	filename - The name of the ini file to create.
 *  Returns:	INI_SUCCESS if the file is successfully created.
 */
int IniCreateFile( String filename )
{
	char	s[ MAX_INI_LINE_LEN ];
	FILE	*file	=NULL;

	if( !filename ) return( INI_INVALID_VALUE );

	file = fopen( filename, "w" );

	if( file )
	{
		sprintf( s, "# NXIni Library - Version %s\n", INI_VERSION_STR );

		IniWriteLine( file, s );
		IniWriteLine( file, "#\n" );

		if( fclose(file) == EOF )
			return( INI_INVALID_FILE );

		return( INI_SUCCESS );
	}

	return( INI_INVALID_FILE );
}

/***********************************************************************/

/*
 *  Proc:	IniReadLine
 *  Callers:
 *  Purpose:	Read the next ENTIRE non-empty line from the file.
 *		Skips comments and blank lines. Returns the file position of
 *		the line of text read.
 *  Params:	file - The file to read from.
 *		keypos - <Return> The file position of the beginning of the line.
 *  Returns:	A pointer to the line of text.  NULL if there is no more text.
 */
String IniReadLine( FILE *file, filepos_t *filepos )
{
	String	line	=NULL;
	char	buf[ MAX_INI_LINE_LEN ];
	Boolean	done	=False;

	if( !file ) return( NULL );

	line = NULL;
	done = False;

	if( filepos ) (*filepos) = ftell( file );

	while( !done && fgets(buf, MAX_INI_LINE_LEN, file) )
	{
		if( buf[ XnEndChar(buf) ] == '\n' ) buf[ XnEndChar(buf) ] = '\0';

		XnCleanString( buf );

		switch( buf[0] )
		{
			case '\0':
			break;

			case '\n':
			break;

			case '#':
			break;

			default:
				line = XnDupString( buf );
				done = True;
			break;
		}

		if( filepos && !done ) (*filepos) = ftell( file );
	}

	return( line );
}

/***********************************************************************/

/*
 *  Proc:	IniReadSectionLine
 *  Callers:
 *  Purpose:	Return the NAME of the next section in the ini file.
 *		Also returns the entire line of that section.
 *  Params:	file - The ini file.
 *		entry - <Return> The entire section entry.
 *  Returns:	The name of the section or NULL if there are no more
 *		sections.
 */
String IniReadSectionLine( FILE *file, String *entry )
{
	String	p	=NULL;
	String	line	=NULL;
	String	sect	=NULL;

	if( !file ) return( NULL );

	while( (line=IniReadLine( file, NULL )) )
	{
		if( *line == '[' )
		{
			if( entry ) (*entry) = XnDupString( line );

			for( p=line + 1; (*p) && (*p != ']'); p++ )
				;
			(*p) = '\0';

			sect = XnDupString( line + 1 );
			XnCleanString( sect );
			XnFreeString( line );

			return( sect );
		}
	}

	return( NULL );
}

/***********************************************************************/

/*
 *  Proc:	IniNextSection
 *  Callers:
 *  Purpose:	Return the NAME (text between brackets) of the next section
 *		in the ini file.
 *  Params:	file - The ini file.
 *  Returns:	The name of the section or NULL if there are no
 *		more sections in the file.
 */
String IniNextSection( FILE *file )
{
	String	sect	=NULL;

	if( !file ) return( NULL );

	sect = IniReadSectionLine( file, (String *)NULL );

	return( sect );
}
/***********************************************************************/

/*
 *  Proc:	IniReadKeyLine
 *  Callers:	
 *  Purpose:	Return the NAME of the next key in the current section.
 *		Also returns the entire line of that section. The file position
 *		is the beginning of the line after the line returned.
 *  Params:	file - The ini filename.
 *		entry - <Return> The key entry.
 *		keypos - <Return> The file position of the returned key.
 *  Returns:	The name of the key or NULL if there are no more keys
 *		in the current section.
 */
String IniReadKeyLine( FILE *file, String *entire, filepos_t *keypos )
{
	String	key	=NULL;
	String	line	=NULL;
	char	buf[ MAX_INI_LINE_LEN ];

	if( !file ) return( NULL );

	while( (line=IniReadLine( file, keypos )) )
	{
		if( *line == '[' )
		{
			XnFreeString( line );
			return( NULL );
		}
		else
		{
			if( entire ) (*entire) = XnDupString( line );

			strcpy( buf, line );
			XnFreeString( line );

			key = XnDupString( strtok( buf, ":" ) );
			if( key )
			{
				XnCleanString( key );
				return( key );
			}
		}
	}

	return( NULL );
}

/***********************************************************************/

/*
 *  Proc:	IniNextKey
 *  Callers:
 *  Purpose:	Return the NAME (not entire line) of the next
 *		key in the current section. The file position is the beginning
 *		of the line after the line returned.
 *  Params:	file - The ini file.
 *		keypos - <Return> The file position of the key returned.
 *  Returns:	The name of the key or NULL if there are no more keys
 *		in the current section.
 */
String IniNextKey( FILE *file, filepos_t *keypos )
{
	String	key	=NULL;

	if( !file ) return( NULL );

	key = IniReadKeyLine( file, NULL, keypos );

	return( key );
}

/***********************************************************************/

/*
 *  Proc:	IniFileExists
 *  Callers:
 *  Purpose:	Determines if the ini file exists.
 *  Params:	filename - The filename of the ini file.
 *  Returns:	True if the file exists.
 */
Boolean IniFileExists( String filename )
{
	return( XnFileExists(filename) );
}

/***********************************************************************/

/*
 *  Proc:	IniSameSection
 *  Callers:
 *  Purpose:	Compares the names of two sections.
 *  Params:	a, b - The section names.
 *  Returns:	0 on equality (line strcmp).
 */
int IniSameSection( String a, String b )
{
	return( strcmp(a, b) );
}

/***********************************************************************/

/*
 *  Proc:	IniSameKey
 *  Callers:
 *  Purpose:	Compares the names of two keys.
 *  Params:	a, b - The key names.
 *  Returns:	0 on equality (line strcmp).
 */
int IniSameKey( String a, String b )
{
	return( strcmp(a, b) );
}

/***********************************************************************/

/*
 *  Proc:	IniSeekToSection
 *  Callers:
 *  Purpose:	Seeks the file to the beginning of the next section
 *		(which is the beginning of the first key in that section).
 *		If section is NULL the file position is set to the
 *		first section of keys in the file.
 *  Params:	file - The ini file.
 *		section - The name of the section.
 *  Returns:	INI_SUCCESS if a section is found.
 */
int IniSeekToSection( FILE *file, String section )
{
	String		s	=NULL;
	filepos_t	fpos	=0;

	if( !file ) return( INI_INVALID_VALUE );

	if( section )
	{
		while( (s=IniNextSection(file)) )
			if( IniSameSection(s, section) == 0 )
				return( INI_SUCCESS );
	}
	else
	{
		/*
		 * Read to the first line in the file.  If it is a section
		 * then continue.  If it is a key, reset the file position
		 * to the beginning of the key.
		 */
		if( (s=IniReadLine(file, &fpos)) )
		{
			if( s[0] == '[' ) return( INI_SUCCESS );
			else
			{
                                if( fseek(file, fpos, SEEK_SET) )
					return( INI_INVALID_FILE );
                                else
                                        return( INI_SUCCESS );
			}
		}
	}

	return( INI_NO_SECTION );
}

/***********************************************************************/

/*
 *  Proc:	IniSeekToKey
 *  Callers:
 *  Purpose:	Seeks the file to the line of the specified key
 *		within the current section.  The file position
 *		is the beginning of the key returned if the key is
 *		found or the beginning of the section following
 *		the specified key if the key is not found in the current
 *		section.  If there are no sections the file position
 *		is undefined.
 *  Params:	file - The ini file.
 *		key - The name of the key.
 *  Returns:	INI_SUCCESS if a key is found.
 */
int IniSeekToKey( FILE *file, String key )
{
	String		s	=NULL;
	filepos_t	keypos	=0;

	if( !file || !key ) return( INI_INVALID_VALUE );

	while( (s=IniNextKey(file, &keypos)) )
	{
		if( IniSameKey(s, key) == 0 )
		{
			if( fseek(file, keypos, SEEK_SET) )
				return( INI_INVALID_FILE );
			else
				return( INI_SUCCESS );
		}
	}

	if( fseek(file, keypos, SEEK_SET) ) return( INI_INVALID_FILE );

	return( INI_NO_KEY );
}

/***********************************************************************/

/*
 *  Proc:	IniSectionExists
 *  Callers:	
 *  Purpose:	Determines if a specified section exists in an ini file.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *  Returns:	True if the section exists.
 */
Boolean IniSectionExists( String filename, String section )
{
	FILE	*file	=NULL;

	if( !filename ) return( False );

	file = fopen( filename, "r" );

	if( file )
	{
		if( IniSeekToSection(file, section) == INI_SUCCESS )
		{
			fclose( file );
			return( True );
		}
		fclose( file );
	}

	return( False );
}

/***********************************************************************/

/*
 *  Proc:	IniSectionKeyExists
 *  Callers:
 *  Purpose:	Determines if a specified key exists within the specified
 *		section.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *  Returns:	True if the section and key exist.
 */
Boolean IniSectionKeyExists( String filename, String section, String key )
{
	FILE	*file	=NULL;

	if( !filename || !key ) return( False );

	file = fopen( filename, "r" );

	if( file )
	{
		if( IniSectionExists(filename, section) )
		{
			if( IniSeekToKey(file, key) == INI_SUCCESS )
			{
				fclose( file );
				return( True );
			}
		}
		fclose( file );
	}

	return( False );
}

/***********************************************************************/

/*
 *  Proc:	IniGetSectionEntry
 *  Callers:
 *  Purpose:	Returns the ENTIRE text entry for the section specified.
 *  Params:	filename - The ini filename.
 *		section - The section to find.
 *  Returns:	The section entry or NULL if it does not exist.
 */
String IniGetSectionEntry( String filename, String section )
{
	String	s	=NULL;
	String	entry	=NULL;
	FILE	*file	=NULL;

	if( !filename || !section ) return( NULL );

	file = fopen( filename, "r" );

	if( file )
	{
		while( (s=IniReadSectionLine( file, &entry )) )
		{
			if( IniSameSection(s, section) == 0 )
			{
				fclose( file );
				XnFreeString( s );

				return( entry );
			}
			XnFreeString( s );
			XnFreeString( entry );
		}
		fclose( file );
	}

	return( NULL );
}

/***********************************************************************/

/*
 *  Proc:	IniGetSectionKeyEntry
 *  Callers:	
 *  Purpose:	Returns the ENTIRE text entry for the key specified.
 *  Params:	filename - The ini filename.
 *		section - The section to find.
 *		key - The key to find.
 *  Returns:	The key entry of NULL if it does not exist.
 */

String IniGetSectionKeyEntry( FILE *file, String section, String key )
{
	String		s	=NULL;
	String		entry	=NULL;
	filepos_t	keypos	=0;

	if( !file || !key ) return( NULL );

	if( IniSeekToSection(file, section) == INI_SUCCESS )
	{
		while( (s=IniReadKeyLine( file, &entry, &keypos )) )
		{
			if( IniSameKey(s, key) == 0 )
			{
				XnFreeString( s );

				return( entry );
			}
			XnFreeString( s );
			XnFreeString( entry );
		}
	}

	return( NULL );
}

/***********************************************************************
 *
 *
 *  The following functions are PUBLIC functions used by applications
 *  to manipulate their configuration files.
 *
 *
 ***********************************************************************/

/*
 *  Proc:	XnCreateIniFilename
 *  Callers:	<Library>
 *  Purpose:	Create an initilization dotfile based on the
 *		name of the application.  The name of the file for an
 *		application named "APP" would be $HOME/.apprc where $HOME
 *		is the users home directory and apprc is all lowercase.
 *  Params:	app - The application name.
 *  Returns:	The complete filename of the ini file.
 */
String  XnCreateIniFilename ( String app )
{
	String			s		=NULL;
	XnUserInfoStruct	*userInfo	=NULL;

	if( !app ) return( NULL );

	userInfo = XnRetrieveUserInfo ( );

	if( !userInfo ) return( NULL );

	s = XnMallocString( strlen(userInfo->homeDir) + 2 + strlen(app) + 2 + 1 );

	sprintf( s, "%s/.%src", userInfo->homeDir, XnLowerString(app) );

	XtFree( (char *)userInfo );

	return( s );
}

/***********************************************************************/

/*
 *  Proc:	XnIniWriteComment
 *  Callers:	<Library>
 *  Purpose:	Appends a comment to the end of the ini file.
 *  Params:	filename - The ini filename.
 *		comment - The comment to write.
 *  Returns:	INI_SUCCESS if successful.
 */
int XnIniWriteComment( String filename, String comment )
{
	int	i;
	char	s[ MAX_INI_LINE_LEN ];
	FILE	*file	=NULL;

	if( !filename || !comment ) return( INI_INVALID_VALUE );

	if( !IniFileExists(filename) )
		IniCreateFile( filename );

	file = fopen( filename, "a" );
	if( file )
	{
		sprintf( s, "#%s\n", comment );

		i = IniWriteLine( file, s );

		fclose( file );
		
		return( i );
	}

	return( INI_INVALID_FILE );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionList
 *  Callers:	<Library>
 *  Purpose:	Return an array of section names that are listed in the
 *		ini file.
 *  Params:	filename - The name of the ini file.
 *		number - <Return> The number of sections returned.
 *  Returns:	An array of section names or NULL if none are defined.
 */
StringTable XnIniGetSectionList ( String filename, int *number )
{
	String		line 	=NULL;
	FILE		*file	=NULL;
	StringTable	list 	=NULL;

	(*number) = 0;

	if( !filename || !number ) return( NULL );

	if( IniFileExists(filename) == False ) return( NULL );

	file = fopen( filename, "r" );
	if( file )
	{
		list = XnMallocStringTable( 1 );

		/* Loop for each section. */
		while( (line=IniNextSection( file )) )
		{
			(*number)++;

			/* Add to the list at the end. */
			list = (StringTable) XtRealloc( (String)list, (Cardinal)(sizeof(String) * (*number)) );
			list[ (*number) - 1 ] = line;
		}

		fclose ( file );

		/* If there were no sections... */
		if( !(*number) )
		{
			XnFreeStringTable( list, 1 );
			list = NULL;
		}

		return( list );
	}

	return( NULL );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyList
 *  Callers:	<Library>
 *  Purpose:	Return an array of keys defined for a particular section.
 *  Params:	filename - The name of the ini file.
 *		section - The section to get the keys for.
 *		number - <Return> The number of keys returned.
 *  Returns:	The list of keys or NULL if none are defined.
 */
StringTable XnIniGetSectionKeyList( String filename, String section, int *number )
{
	String		line 	=NULL;
	FILE		 *file	=NULL;
	StringTable	list 	=NULL;

	(*number) = 0;

	if( !filename || !number ) return( NULL );

	if( IniFileExists(filename) == False ) return( NULL );

	file = fopen( filename, "r" );
	if( file )
	{
		if( IniSeekToSection(file, section) == INI_SUCCESS )
		{
			list = XnMallocStringTable( 1 );

			/* Loop for each key. */
			while( (line=IniNextKey(file, NULL)) )
			{
				(*number)++;

				/* Add to the list at the end. */
				list = (StringTable) XtRealloc( (String)list, (Cardinal)(sizeof(String) * (*number)) );
				list[ (*number) - 1 ] = line;
			}

			fclose ( file );

			/* If there were no keys... */
			if( !(*number) )
			{
				XnFreeStringTable( list, 1 );
				list = NULL;
			}

			return( list );
		}
	}

	return( NULL );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyString
 *  Callers:	<Library>
 *  Purpose:	Returns the string value of a specified section/key combination.
 *  Params:	filename - The name of the ini file.
 *		section - The section the key is under.
 *		key - The key to search for.
 *		retValue - Where to store the value.
 *		defaultValue - The default value to use if the key does not exist.
 *  Returns:	INI_SUCCESS if a key is found.  INI_NO_KEY is returned if the
 *		key is not found, but the default value is still used.
 */
int XnIniGetSectionKeyString ( String filename, String section, String key, String *retValue, String defaultValue )
{
	FILE		*file		=NULL;
	String		line		=NULL;

	if( !filename || !key || !retValue ) return( INI_INVALID_VALUE );

	if( IniFileExists( filename ) == False ) return( INI_INVALID_FILE );

	file = fopen( filename, "r" );
	if( file )
	{
		/*
		 * Try to find the key.  If it can't be found use the
		 * default value (if one is specified).
 		 */
		if( (line=IniGetSectionKeyEntry( file, section, key )) )
		{
			fclose( file );

			strtok( line, ":" );
			(*retValue) = XnDupString( strtok( NULL, "\n" ) );
			XnCleanString( (*retValue) );

			XnFreeString( line );

			return( INI_SUCCESS );
		}
		else
		{
			fclose ( file );

			if( defaultValue ) (*retValue) = XnDupString( defaultValue );

			return( INI_NO_KEY );
		}
	}

	return( INI_INVALID_FILE );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyStringList
 *  Callers:	<Library>
 *  Purpose:	Return the string list value of a section/key combination.
 *  Params:	filename - The name of the ini file.
 *		section - The section the key is under.
 *		key - The key to search for.
 *		retValue - Where to store the value.
 *		defaultValue - The default value to use if the key does not exist.
 *  Returns:	INI_SUCCESS if a key is found.  INI_NO_KEY is returned if the
 *		key is not found, but the default value is still used.
 */
int XnIniGetSectionKeyStringList( String filename, String section, String key, StringTable *retValue, StringTable *defaultValue )
{
	int	i;
	int	iniVal		=0;
	String	strval		=NULL;

	if( !filename || !key || !retValue ) return( INI_INVALID_VALUE );

	iniVal = XnIniGetSectionKeyString( filename, section, key, &strval, NULL );

	/*
	 * Get the value associated with the key as a string.  If the key is found
	 * then convert the value to a string list. Use the default value if needed.
	 */
	if( iniVal == INI_SUCCESS )
	{
		(*retValue) = XnSplitString( strval, ",", &i, False );
	}

	else if( iniVal == INI_NO_KEY )
	{
		if( defaultValue != NULL )
			(*retValue) = (*defaultValue);
	}

	return( iniVal );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyInt
 *  Callers:	<Library>
 *  Purpose:	Return the integer value of a section/key combination.
 *  Params:	filename - The name of the ini file.
 *		section - The section the key is under.
 *		key - The key to search for.
 *		retValue - Where to store the value.
 *		defaultValue - The default value to use if the key does not exist.
 *  Returns:	INI_SUCCESS if a key is found.  INI_NO_KEY is returned if the
 *		key is not found, but the default value is still used.
 */
int XnIniGetSectionKeyInt ( String filename, String section, String key, int *retValue, int *defaultValue )
{
	int	iniVal		=0;
	String	strval		=NULL;

	if( !filename || !key || !retValue ) return( INI_INVALID_VALUE );

	iniVal = XnIniGetSectionKeyString( filename, section, key, &strval, NULL );

	/*
	 * Get the value associated with the key as a string.  If the key is found
	 * then convert the value to an integer. Use the default value if needed.
	 */
	if( iniVal == INI_SUCCESS )
	{
		if( strval )
			(*retValue) = atoi( strval );
		else
			(*retValue) = 0;
	}

	else if( iniVal == INI_NO_KEY )
	{
		if( defaultValue != NULL )
			(*retValue) = (*defaultValue);
	}

	return( iniVal );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyFloat
 *  Callers:	<Library>
 *  Purpose:	Return the floating point value of a section/key combination.
 *  Params:	filename - The name of the ini file.
 *		section - The section the key is under.
 *		key - The key to search for.
 *		retValue - Where to store the value.
 *		defaultValue - The default value to use if the key does not exist.
 *  Returns:	INI_SUCCESS if a key is found.  INI_NO_KEY is returned if the
 *		key is not found, but the default value is still used.
 */
int XnIniGetSectionKeyFloat( String filename, String section, String key, float *retValue, float *defaultValue )
{
	int	iniVal		=0;
	String	strval		=NULL;

	if( !filename || !key || !retValue ) return( INI_INVALID_VALUE );

	iniVal = XnIniGetSectionKeyString( filename, section, key, &strval, NULL );

	/*
	 * Get the value associated with the key as a string.  If the key is found
	 * then convert the value to an integer. Use the default value if needed.
	 */
	if( iniVal == INI_SUCCESS )
	{
		if( strval )
			(*retValue) = (float)atof( strval );
		else
			(*retValue) = (float)0.0;
	}

	else if( iniVal == INI_NO_KEY )
	{
		if( defaultValue != NULL )
			(*retValue) = (float)(*defaultValue);
	}

	return( iniVal );
}

/***********************************************************************/

/*
 *  Proc:	XnIniGetSectionKeyBoolean
 *  Callers:	<Library>
 *  Purpose:	Return the boolean value of a section/key combination.
 *  Params:	filename - The name of the ini file.
 *		section - The section the key is under.
 *		key - The key to search for.
 *		retValue - Where to store the value.
 *		defaultValue - The default value to use if the key does not exist.
 *  Returns:	INI_SUCCESS if a key is found.  INI_NO_KEY is returned if the
 *		key is not found, but the default value is still used.
 */
int XnIniGetSectionKeyBoolean ( String filename, String section, String key, Boolean *retValue, Boolean *defaultValue )
{
	int	iniVal		=0;
	String	strval		=NULL;

	if( !filename || !key || !retValue ) return( INI_INVALID_VALUE );

	iniVal = XnIniGetSectionKeyString( filename, section, key, &strval, NULL );

	/*
	 * Get the value associated with the key as a string.  If the key is found
	 * then convert the value to an integer. Use the default value if needed.
	 */
	if( iniVal == INI_SUCCESS )
	{
		if( strval )
		{
			if( (strcmp(XnLowerString(strval), "true") == 0) || (strcmp(XnLowerString(strval), "1") == 0) )
				(*retValue) = True;
			else
				(*retValue) = False;
		}
		else (*retValue) = False;
	}

	else if( iniVal == INI_NO_KEY )
	{
		if( defaultValue != NULL )
			(*retValue) = (*defaultValue);
	}

	return( iniVal );
}

/***********************************************************************/

/*
 *  Proc:	XnIniLoadResources
 *  Callers:	<Library>
 *  Purpose:	Loads an entire XtResourceList at once.  See the include file
 *		NXUtils.h for a more complete description.
 *  Params:	filename - The name of the ini file.
 *		base - A pointer to the base of a storage structure.
 *		resources - The resources to load.
 *		num_resources - The number of resources to load.
 *  Returns:	True if successful.
 */
Boolean XnIniLoadResources( String filename, XtPointer base, XtResourceList resources, Cardinal num_resources )
{
	int		i;
	int		v;
	unsigned int	addr		=0;

	if( !filename || !IniFileExists(filename) ) return( False );

	v = INI_SUCCESS;

	for( i=0; i < num_resources; i++ )
	{
		addr = (unsigned int)base + (unsigned int)resources[i].resource_offset;

		if( strcmp( (char *)resources[i].resource_type, (char *)XtRString ) == 0 )
		{
			v |= XnIniGetSectionKeyString( filename,
					resources[i].resource_class,
					resources[i].resource_name,
					(StringTable)addr,
					(String)resources[i].default_addr );
		}

		else if( strcmp( (char *)resources[i].resource_type, (char *)XtRFloat ) == 0 )
		{
			v |= XnIniGetSectionKeyFloat( filename,
					resources[i].resource_class,
					resources[i].resource_name,
					(float *)addr,
					(float *)resources[i].default_addr );
		}

		else if( strcmp( (char *)resources[i].resource_type, (char *)XtRInt ) == 0 )
		{
			v |= XnIniGetSectionKeyInt( filename,
					resources[i].resource_class,
					resources[i].resource_name,
					(int *)addr,
					(int *)resources[i].default_addr );
		}

		else if( strcmp( (char *)resources[i].resource_type, (char *)XtRBoolean ) == 0 )
		{
			v |= XnIniGetSectionKeyBoolean( filename,
					resources[i].resource_class,
					resources[i].resource_name,
					(Boolean *)addr,
					(Boolean *)resources[i].default_addr );
		}
	}

	if( (v == INI_SUCCESS) || (v == INI_NO_KEY) ) return( True );

	return( False );
}

/***********************************************************************/

/*
 *  Proc:	XnIniSetSectionKeyString
 *  Callers:	<Library>
 *  Purpose:	Set/Change the string value of a section/key combination.  If the section
 *		or key do not exist they will be created.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *		newValue - The new value of the section/key.
 *  Returns:	INI_SUCCESS if the value was successfully set.
 */
int XnIniSetSectionKeyString ( String filename, String section, String key, String newValue )
{
	char		s[ MAX_INI_LINE_LEN ];
	String		textbuf	=NULL;
	FILE		*file 		=NULL;
	filepos_t	pos 		=0;
	Boolean		newFile		=False;

	if( !filename || !key || !newValue ) return( INI_INVALID_VALUE );

	/*
	 * If the file dosen't exist - create it.
	 */
	if( IniFileExists(filename) == False )
	{
		newFile = True;

		if( IniCreateFile( filename ) != INI_SUCCESS )
			return( INI_INVALID_FILE );

		if( (file=fopen( filename, "a" )) == NULL )
			return( INI_INVALID_FILE );
	}
	else
	{
		newFile = False;
	
		if( (file=fopen( filename, "r+" )) == NULL )
			return( INI_INVALID_FILE );
	}

	clearerr( file );

	/*
	 * First check to see if the section exists.  If the section
	 * exists we have to do some extra work - otherwise we just
	 * add the new section and key to the end of the file.
	 */
	if( !newFile && (IniSeekToSection( file, section ) == INI_SUCCESS) )
	{
		/*
		 * Now check to see if the key already exists.
		 * The file position will be the beginning of the key
		 * returned if the key is found or the beginning
		 * of the section following the specified key if
		 * the key is not found in the current section.
		 * If the key exists we save the entry line of the key.
		 * In either case, remember where we are in the file so
		 * we can put the file back together.
		 */
		if ( IniSeekToKey ( file, key ) == INI_SUCCESS )
		{
			pos = ftell( file );
			fgets( s, MAX_INI_LINE_LEN, file );
		}
		else pos = ftell( file );

		textbuf = XnMallocString( 2 );
		strcpy ( textbuf, "" );

		/*
		 * Algorithm:  The text in the file UP TO the specified
		 * key (or next section) is in the file.  This position
		 * was saved by ftell.  We already deleted the line
		 * with the key.  Now we take the rest of the text in the
		 * file (after the key or section) and save it in a buffer.
		 * This is basically just splicing the new key value into
		 * the file.
		 */
		while( fgets( s, MAX_INI_LINE_LEN, file ) )
		{
			int len = strlen ( textbuf );

			textbuf = ( String ) XtRealloc ( textbuf, ( len + strlen (s) + 2 ) );

			strcat( textbuf, s );
		}

		/*
		 * Reset the file position, output the new key value,
		 * then output the rest of the text.
		 */
		if( fseek( file, pos, SEEK_SET ) == 0 )
		{
			sprintf( s, "%s : %s\n", key, newValue );

			IniWriteLine( file, s );
			IniWriteLine( file, textbuf );
		}
	}

	/*
	 * If it is a new file or if the section didn't exist we
	 * just output the new section and key values.
	 */
	else
	{
		IniWriteLine( file, "\n" );

		sprintf( s, "[ %s ]\n\n", section );
		IniWriteLine( file, s );

		sprintf( s, "%s : %s\n", key, newValue );
		IniWriteLine( file, s );
	}

	if( (fflush(file) == EOF) || (fclose(file) == EOF) )
		return( INI_INVALID_FILE );

	if( ferror(file) == 0 )
		return( INI_SUCCESS );

	return( INI_INVALID_FILE );
}

/***********************************************************************/

/*
 *  Proc:	XnIniSetSectionKeyInt
 *  Callers:	<Library>
 *  Purpose:	Set/Change the integer value of a section/key combination.  If the section
 *		or key do not exist they will be created.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *		newValue - The new value of the section/key.
 *  Returns:	INI_SUCCESS if the value was successfully set.
 */
int XnIniSetSectionKeyInt ( String filename, String section, String key, int newValue )
{
	char	s[ MAX_INI_LINE_LEN ];

	sprintf ( s, "%d", newValue );

	return( XnIniSetSectionKeyString ( filename, section, key, s ) );
}

/***********************************************************************/

/*
 *  Proc:	XnIniSetSectionKeyFloat
 *  Callers:	<Library>
 *  Purpose:	Set/Change the floating point value of a section/key combination.  If the section
 *		or key do not exist they will be created.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *		newValue - The new value of the section/key.
 *  Returns:	INI_SUCCESS if the value was successfully set.
 */
int XnIniSetSectionKeyFloat ( String filename, String section, String key, float newValue )
{
	char	s[ MAX_INI_LINE_LEN ];

	sprintf ( s, "%f", newValue );

	return( XnIniSetSectionKeyString ( filename, section, key, s ) );
}

/***********************************************************************/

/*
 *  Proc:	XnIniSetSectionKeyBoolean
 *  Callers:	<Library>
 *  Purpose:	Set/Change the boolean value of a section/key combination.  If the section
 *		or key do not exist they will be created.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *		newValue - The new value of the section/key.
 *  Returns:	INI_SUCCESS if the value was successfully set.
 */
int XnIniSetSectionKeyBoolean ( String filename, String section, String key, Boolean newValue )
{
	if( newValue )
		return( XnIniSetSectionKeyString ( filename, section, key, "True" ) );
	else
		return( XnIniSetSectionKeyString ( filename, section, key, "False" ) );
}

/***********************************************************************/

/*
 *  Proc:	XnIniSetSectionKeyStringList
 *  Callers:	<Library>
 *  Purpose:	Set/Change the string list value of a section/key combination.  If the section
 *		or key do not exist they will be created.
 *  Params:	filename - The ini filename.
 *		section - The name of the section.
 *		key - The name of the key.
 *		newValue - The new value of the section/key.
 *  Returns:	INI_SUCCESS if the value was successfully set.
 */
int XnIniSetSectionKeyStringList( String filename, String section, String key, String *newValue )
{
	int	i;
	char	s[ MAX_INI_LINE_LEN ];

	if( newValue )
	{
		strcpy( s, newValue[0] );

		for( i=1; newValue[i]; i++ )
		{
			strcat( s, "," );
			strcat( s, newValue[i] );
		}
	}
	else strcpy( s, "" );

	return( XnIniSetSectionKeyString ( filename, section, key, s ) );
}

